@@ -1,4 +1,4 @@
-#!/usr/bin/env python
+#!/usr/bin/env python3
 # Copyright 2015 gRPC authors.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
@@ -27,13 +27,15 @@ import subprocess
 import sys
 import tempfile
 import time
-import uuid
-import six
 import traceback
+import uuid
+
+import six
 
 import python_utils.dockerjob as dockerjob
 import python_utils.jobset as jobset
 import python_utils.report_utils as report_utils
+
 # It's ok to not import because this is only necessary to upload results to BQ.
 try:
     from python_utils.upload_test_results import upload_interop_results_to_bq
@@ -158,8 +160,8 @@ class CSharpLanguage:
 class CSharpCoreCLRLanguage:
 
     def __init__(self):
-        self.client_cwd = 'src/csharp/Grpc.IntegrationTesting.Client/bin/Debug/netcoreapp2.1'
-        self.server_cwd = 'src/csharp/Grpc.IntegrationTesting.Server/bin/Debug/netcoreapp2.1'
+        self.client_cwd = 'src/csharp/Grpc.IntegrationTesting.Client/bin/Debug/netcoreapp3.1'
+        self.server_cwd = 'src/csharp/Grpc.IntegrationTesting.Server/bin/Debug/netcoreapp3.1'
         self.safename = str(self)
 
     def client_cmd(self, args):
@@ -487,6 +489,7 @@ class PHP7Language:
 
     def __init__(self):
         self.client_cwd = None
+        self.server_cwd = None
         self.safename = str(self)
 
     def client_cmd(self, args):
@@ -495,6 +498,9 @@ class PHP7Language:
     def cloud_to_prod_env(self):
         return {}
 
+    def server_cmd(self, args):
+        return ['src/php/bin/interop_server.sh'] + args
+
     def global_env(self):
         return {}
 
@@ -505,7 +511,7 @@ class PHP7Language:
             _SKIP_COMPUTE_ENGINE_CHANNEL_CREDS
 
     def unimplemented_test_cases_server(self):
-        return []
+        return _SKIP_COMPRESSION
 
     def __str__(self):
         return 'php7'
@@ -590,6 +596,9 @@ class RubyLanguage:
         return 'ruby'
 
 
+_PYTHON_BINARY = 'py39/bin/python'
+
+
 class PythonLanguage:
 
     def __init__(self):
@@ -600,13 +609,13 @@ class PythonLanguage:
 
     def client_cmd(self, args):
         return [
-            'py37_native/bin/python', 'src/python/grpcio_tests/setup.py',
-            'run_interop', '--client', '--args="{}"'.format(' '.join(args))
+            _PYTHON_BINARY, 'src/python/grpcio_tests/setup.py', 'run_interop',
+            '--client', '--args="{}"'.format(' '.join(args))
         ]
 
     def client_cmd_http2interop(self, args):
         return [
-            'py37_native/bin/python',
+            _PYTHON_BINARY,
             'src/python/grpcio_tests/tests/http2/negative_http2_client.py',
         ] + args
 
@@ -615,8 +624,8 @@ class PythonLanguage:
 
     def server_cmd(self, args):
         return [
-            'py37_native/bin/python', 'src/python/grpcio_tests/setup.py',
-            'run_interop', '--server', '--args="{}"'.format(' '.join(args))
+            _PYTHON_BINARY, 'src/python/grpcio_tests/setup.py', 'run_interop',
+            '--server', '--args="{}"'.format(' '.join(args))
         ]
 
     def global_env(self):
@@ -648,14 +657,13 @@ class PythonAsyncIOLanguage:
 
     def client_cmd(self, args):
         return [
-            'py37_native/bin/python', 'src/python/grpcio_tests/setup.py',
-            'run_interop', '--use-asyncio', '--client',
-            '--args="{}"'.format(' '.join(args))
+            _PYTHON_BINARY, 'src/python/grpcio_tests/setup.py', 'run_interop',
+            '--use-asyncio', '--client', '--args="{}"'.format(' '.join(args))
         ]
 
     def client_cmd_http2interop(self, args):
         return [
-            'py37_native/bin/python',
+            _PYTHON_BINARY,
             'src/python/grpcio_tests/tests/http2/negative_http2_client.py',
         ] + args
 
@@ -664,8 +672,8 @@ class PythonAsyncIOLanguage:
 
     def server_cmd(self, args):
         return [
-            'py37_native/bin/python', 'src/python/grpcio_tests/setup.py',
-            'run_interop', '--use-asyncio', '--server',
+            _PYTHON_BINARY, 'src/python/grpcio_tests/setup.py',
+            'py39/bin/python', 'src/python/grpcio_tests/setup.py',
             '--args="{}"'.format(' '.join(args))
         ]
 
@@ -714,7 +722,7 @@ _LANGUAGES = {
 # languages supported as cloud_to_cloud servers
 _SERVERS = [
     'c++', 'node', 'csharp', 'csharpcoreclr', 'aspnetcore', 'java', 'go',
-    'ruby', 'python', 'dart', 'pythonasyncio'
+    'ruby', 'python', 'dart', 'pythonasyncio', 'php7'
 ]
 
 _TEST_CASES = [
@@ -748,8 +756,8 @@ _GRPC_CLIENT_TEST_CASES_FOR_HTTP2_SERVER_TEST_CASES = {
     'no_df_padding_sanity_test': 'large_unary'
 }
 
-_HTTP2_SERVER_TEST_CASES_THAT_USE_GRPC_CLIENTS = _GRPC_CLIENT_TEST_CASES_FOR_HTTP2_SERVER_TEST_CASES.keys(
-)
+_HTTP2_SERVER_TEST_CASES_THAT_USE_GRPC_CLIENTS = list(
+    _GRPC_CLIENT_TEST_CASES_FOR_HTTP2_SERVER_TEST_CASES.keys())
 
 _LANGUAGES_WITH_HTTP2_CLIENTS_FOR_HTTP2_SERVER_TEST_CASES = [
     'java', 'go', 'python', 'c++'
@@ -770,11 +778,15 @@ DOCKER_WORKDIR_ROOT = '/var/local/git/grpc'
 
 def docker_run_cmdline(cmdline, image, docker_args=[], cwd=None, environ=None):
     """Wraps given cmdline array to create 'docker run' cmdline from it."""
-    docker_cmdline = ['docker', 'run', '-i', '--rm=true']
+
+    docker_cmdline = ['docker', 'run', '--rm=true']
+    if sys.stdout.isatty():
+        # use "-it" when TTY is available to allow Ctrl-C to work
+        docker_cmdline.append('-it')
 
     # turn environ into -e docker args
     if environ:
-        for k, v in environ.items():
+        for k, v in list(environ.items()):
             docker_cmdline += ['-e', '%s=%s' % (k, v)]
 
     # set working directory
@@ -1128,8 +1140,6 @@ def build_interop_image_jobspec(language, tag=None):
         'INTEROP_IMAGE': tag,
         'BASE_NAME': 'grpc_interop_%s' % language.safename
     }
-    if not args.travis:
-        env['TTY_FLAG'] = '-t'
     build_job = jobset.JobSpec(
         cmdline=['tools/run_tests/dockerize/build_interop_image.sh'],
         environ=env,
@@ -1200,7 +1210,7 @@ argp.add_argument('--google_default_creds_use_key_file',
                         'google_default_credentials test case, e.g. by '
                         'setting env var GOOGLE_APPLICATION_CREDENTIALS.'))
 argp.add_argument('--prod_servers',
-                  choices=prod_servers.keys(),
+                  choices=list(prod_servers.keys()),
                   default=['default'],
                   nargs='+',
                   help=('The servers to run cloud_to_prod and '
@@ -1231,11 +1241,14 @@ argp.add_argument(
     type=str,
     help='Default GCE service account email to use for some auth interop tests.',
     default='830293263384-compute@developer.gserviceaccount.com')
-argp.add_argument('-t',
-                  '--travis',
-                  default=False,
-                  action='store_const',
-                  const=True)
+argp.add_argument(
+    '-t',
+    '--travis',
+    default=False,
+    action='store_const',
+    const=True,
+    help='When set, indicates that the script is running on CI (= not locally).'
+)
 argp.add_argument('-v',
                   '--verbose',
                   default=False,
@@ -1541,7 +1554,7 @@ try:
         (server_host, server_port) = server[1].split(':')
         server_addresses[server_name] = (server_host, server_port)
 
-    for server_name, server_address in server_addresses.items():
+    for server_name, server_address in list(server_addresses.items()):
         (server_host, server_port) = server_address
         server_language = _LANGUAGES.get(server_name, None)
         skip_server = []  # test cases unimplemented by server
@@ -1655,7 +1668,7 @@ try:
 
     report_utils.render_junit_xml_report(resultset, _TESTS_XML_REPORT)
 
-    for name, job in resultset.items():
+    for name, job in list(resultset.items()):
         if "http2" in name:
             job[0].http2results = aggregate_http2_results(job[0].message)
 
@@ -1668,7 +1681,7 @@ try:
         sys.exit(0)
 finally:
     # Check if servers are still running.
-    for server, job in server_jobs.items():
+    for server, job in list(server_jobs.items()):
         if not job.is_running():
             print('Server "%s" has exited prematurely.' % server)
 
